/*
** $Id: mod_video.c,v 1.5 2000/09/09 02:04:55 borud Exp $
**
** Copyright (C) 2000 Bjrn Borud <borud@pvv.org>
**
** this software is released under the GNU GENERAL PUBLIC LICENSE
** Version 2.  please read the accompanying file named 'COPYING'
** to learn more about your rights and obligations concerning
** this software.
**
** mod_video
**   ...is an Apache module for serving snapshots from the
**   video4linux API.  this code is currently pretty experimental
**   since I'm not all that familiar with the video4linux API.
**   the source is released in the hope that it might inspire
**   someone to improve it so it can be more useful to more people.
**
**
** (the following is used by Apache to make sure we link with the
**  appropriate libs)

MODULE-DEFINITION-START
Name: video_module
ConfigStart
  LIBS="$LIBS -lz -lpng -ljpeg"
  echo " + using libz for compression support"
  echo " + using libpng for png support"
  echo " + using libjpeg for jpeg support"
ConfigEnd
MODULE-DEFINITION-END

*/


#include <sys/types.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#include <png.h>
#include <jpeglib.h>

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"

/* configurable defaults */
#define VIDEO_DEFAULT_DEVICE_NAME "/dev/video"
#define VIDEO_DEFAULT_FORMAT      "jpeg"
#define VIDEO_DEFAULT_INPUT         2  /* channel 2 */
#define VIDEO_DEFAULT_WIDTH       640
#define VIDEO_DEFAULT_HEIGHT      480
#define VIDEO_DEFAULT_MAX_RETRY    10

/* per directory configuration structure */
typedef struct {
    char *device_name;
    char *format;
    int  input;
    int  width;
    int  height;
    int  jpeg_quality;
} video_dir_config;

/* hack to make PNG libs happy */
static request_rec *rglobal;

/* forward declaration of video_module */
module MODULE_VAR_EXPORT video_module;

/* {{{ select_video_channel */

static int
select_video_channel (int channel, int device)
{
    struct video_channel vid_chnl; 

    vid_chnl.channel = channel;

    /* 
    ** apparently I am supposed to query the video channel before I
    ** can select it.  v4l docs suck.
    */
    if (ioctl (device, VIDIOCGCHAN, &vid_chnl) == -1) {
	ap_log_error(APLOG_MARK, APLOG_ERR, NULL,
		     "ioctl() VIDIOCGCHAN, unable to query channel %d", channel);
	return -1;
    }

    /* now we can select channels */
    if (ioctl (device, VIDIOCSCHAN, &vid_chnl) == -1) {
	ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
		     "Failed to select video channel %d: %s", 
		     channel,
		     strerror(errno));
	return -1;
    }

    return 0;
}

/* }}} */

/* {{{ print_raw_image */

static int 
print_raw_image (char *buf, int width, int height, int size, request_rec *r)
{
    int len;

    /* set appropriate MIME type and send image */
    r->content_type = "image/x-rgb";
    ap_send_http_header(r);

    len = ap_rwrite(buf, size, r);
    if (len < size) {
	return -1;
    }

    return 0;
}

/* }}} */
/* {{{ print_jpeg_image */

static int
print_jpeg_image (char *buf, int width, int height, int size, request_rec *r)
{
    int y, x, line_width;
    JSAMPROW row_ptr[1];
    struct jpeg_compress_struct cjpeg;
    struct jpeg_error_mgr jerr;
    char *line;
    void *dconf;
    video_dir_config *conf;
    FILE *tmpfile;
    char *filename;
    int retval = 0;
    struct stat st;

    dconf = r->per_dir_config;
    conf = (video_dir_config *) ap_get_module_config(dconf, &video_module);

    /* open the temporary file */
    filename = tempnam("/tmp", "video");
    if (!filename) {
	retval = -1;
	goto jpeg_cleanup;
    }

    tmpfile = ap_pfopen(r->pool, filename, "w+");
    if (! tmpfile) {
	retval = -1;
	goto jpeg_cleanup;

    }

    line = ap_palloc(r->pool, width * 3);
    if (!line) {
	retval = -1;
	goto jpeg_cleanup;
    }

    /* error handling needs to be done a lot better */
    cjpeg.err = jpeg_std_error(&jerr);
    jpeg_create_compress (&cjpeg);

    /* set image parameters */
    cjpeg.image_width = width;
    cjpeg.image_height= height;
    cjpeg.input_components = 3;
    cjpeg.in_color_space = JCS_RGB;
    jpeg_set_defaults (&cjpeg);
    
    jpeg_set_quality (&cjpeg, conf->jpeg_quality, TRUE);
    cjpeg.dct_method = JDCT_FASTEST;  /* should be configurable */
    jpeg_stdio_dest (&cjpeg, tmpfile);

    jpeg_start_compress (&cjpeg, TRUE);
    
    row_ptr[0] = line;
    line_width = width * 3;
    for ( y = 0; y < height; y++) {
	for (x = 0; x < line_width; x+=3) {
	    line[x]   = buf[x+2];
	    line[x+1] = buf[x+1];
	    line[x+2] = buf[x];
	}
	jpeg_write_scanlines (&cjpeg, row_ptr, 1);
	buf += line_width;
    }
    jpeg_finish_compress (&cjpeg);
    jpeg_destroy_compress (&cjpeg);

    /* rewind fd */
    fseek(tmpfile, 0L, SEEK_SET);

    /* set appropriate MIME type and send image */
    r->content_type = "image/jpeg";

    if (stat(filename, &st)) {
	ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
		     "stat(): %s", strerror(errno));
    } else {
	ap_set_content_length(r, st.st_size);
    }

    /* send headers */
    ap_send_http_header(r);

    /* and send the file */
    ap_send_fd(tmpfile, r);

 jpeg_cleanup:
    if (filename) {
	if (tmpfile) {
	    ap_pfclose(r->pool, tmpfile);
	}

	if (unlink(filename)) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
			 "unlink(): %s", strerror(errno));
	}
	free(filename);
    }

    return retval;
}

/* }}} */
/* {{{ print_png_image */

static void
video_png_error_fn (png_structp png_ptr, png_const_charp error_msg)
{
    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
		 "libpng ERROR: %s", (char *) error_msg);    

}


static void
video_png_warning_fn (png_structp png_ptr, png_const_charp error_msg)
{
    ap_log_error(APLOG_MARK, APLOG_WARNING, NULL, 
		 "libpng WARNING: %s", (char *) error_msg);    
}

static void
video_png_write_data (png_structp png_ptr, png_bytep data, png_size_t length)
{
    png_uint_32 check;

    check = ap_rwrite(data, length, rglobal);
    if (check != length)
    {
	png_error(png_ptr, "Write Error");
    }
}

static void 
video_png_flush_data (png_structp png_ptr)
{
    /* no op */
    return;
}


static int
print_png_image (char *buf, int width, int height, int size, request_rec *r)
{
    png_infop info_ptr;
    png_structp png_ptr;
    int y;
    char *p;

    rglobal = r;

    /* set appropriate MIME type and send image */
    r->content_type = "image/png";
    ap_send_http_header(r); 

    png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 
				       NULL, NULL, NULL);
    if (!png_ptr) {
	return -1;
    }

    info_ptr = png_create_info_struct (png_ptr);
    if (!info_ptr) {
	return -1;
    }

    /* hope this works... */
    png_set_error_fn(png_ptr,
		     NULL,
		     video_png_error_fn,
		     video_png_warning_fn);
		     

    png_set_write_fn(png_ptr, 
		     NULL, 
		     video_png_write_data, 
		     video_png_flush_data);

    png_set_IHDR (png_ptr, info_ptr, width, height,
		  8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		  PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    png_set_compression_level(png_ptr, 6);

    png_set_bgr (png_ptr);
    png_write_info (png_ptr, info_ptr);
    p = buf;
    for (y = 0; y < height; y++) {
	png_write_row (png_ptr, p);
	p+=width*3;
    }
    png_write_end (png_ptr, info_ptr);

    return 0;
}

/* }}} */

/* {{{ video_init */

static void 
video_init (server_rec *s, pool *p)
{
    /* do nothing? */
}

/* }}} */
/* {{{ video_handler */

static int 
video_handler (request_rec *r)
{
    void *dconf;
    int device;
    video_dir_config *conf;
    struct video_capability vid_caps;
    struct video_mbuf vid_buf;
    struct video_mmap vid_mmap;
    char *map;
    int size = 0;
    int tmpsize = 0;
    int len  = 0;
    int attempts = 0;

    /* 
    ** XXX: hmm, we should say something meaningful in the header. 
    **      at least we should present a correct content-type
    */
    if(r->header_only) {
	ap_send_http_header(r); 
	ap_kill_timeout(r);
	return OK;
    }

    /* get config data */
    dconf = r->per_dir_config;
    conf = (video_dir_config *) ap_get_module_config(dconf, &video_module);

    /*
    ** open the video device file.  if we get an EBUSY we try again
    ** a bit later.
    */
    while (attempts < VIDEO_DEFAULT_MAX_RETRY) {
	device = open(conf->device_name, O_RDWR);
	if (device != -1) {
	    break;
	}

	if (errno == EBUSY) {
	    sleep(1);
	} else {
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
			 "Failed to open video device");
	    return SERVER_ERROR;
	}

	attempts++;
    }

    /* warn if max attempts reached */
    if (attempts >= VIDEO_DEFAULT_MAX_RETRY) {
	ap_send_http_header(r); 
	ap_rprintf(r, "Max number of attempts reached.  giving up\n");
	return OK;
    }

    /* select correct video channel */
    if (select_video_channel(conf->input, device)) {
	close(device);
	return SERVER_ERROR;
    }

    /* check if we can use read() */
    if (ioctl (device, VIDIOCGMBUF, &vid_buf) == -1) {
	map = ap_palloc(r->pool, (conf->width * conf->height * 3));
	len = read(device, map, (conf->width * conf->height * 3));
	if (len <= 0) {
	    size = 0;
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, "read() error");	    
	    close(device);
	    return SERVER_ERROR;
	}
    } 

    /* if we can't use read we mmap() the frame */
    else {


	map = mmap (0, vid_buf.size, PROT_READ|PROT_WRITE,MAP_SHARED,device,0);
	if ((unsigned char *)map == (unsigned char *)-1) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, "mmap() failed");
	    close(device);
	    return SERVER_ERROR;
	}
	
	/* set options for capture */
	vid_mmap.format = VIDEO_PALETTE_RGB24;
	vid_mmap.width  = conf->width;
	vid_mmap.height = conf->height;
	vid_mmap.frame  = 0;
	
	/* capture to memory buffer */
	if (ioctl (device, VIDIOCMCAPTURE, &vid_mmap) == -1) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
			 "ioctl() VIDIOCMCAPTURE error");
	    munmap (map, vid_buf.size);
	close(device);
	return SERVER_ERROR;
	}
	
	/* 
	** free up frame (get some more docs on this. the v4l API docs 
	** suck badly)
	*/
	if (ioctl (device, VIDIOCSYNC, &vid_mmap) == -1) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, NULL, 
			 "ioctl() VIDIOCSYNC error");
	    munmap (map, vid_buf.size);
	    close(device);
	    return SERVER_ERROR;
	}

	close(device);
	size = vid_buf.size;
    }

    /* do something to "map" */
    tmpsize = size?size:(conf->width * conf->height * 3);

    if (strncmp(conf->format, "png", 3) == 0) {
	print_png_image (map, conf->width, conf->height, tmpsize, r);
    } 

    else if (strncmp(conf->format, "raw", 3) == 0) {
	print_raw_image (map, conf->width, conf->height, tmpsize, r);
    } 
    
    else {
	print_jpeg_image (map, conf->width, conf->height, tmpsize, r);	    
    }

    
    /* clean up memory mappings */
    if (size) {
	munmap (map, size);
    }
    
    return OK;
}

/* }}} */
/* {{{ video_create_dir_config */

static void *
video_create_dir_config(pool *p, char *dirspec)
{
    video_dir_config *vconf;

    vconf = (video_dir_config *) ap_palloc(p, sizeof(video_dir_config));
    if (! vconf) {
	return NULL;
    }

    vconf->device_name = VIDEO_DEFAULT_DEVICE_NAME;
    vconf->format      = VIDEO_DEFAULT_FORMAT;
    vconf->input       = VIDEO_DEFAULT_INPUT;
    vconf->width       = VIDEO_DEFAULT_WIDTH;
    vconf->height      = VIDEO_DEFAULT_HEIGHT;
}

/* }}} */

/* {{{ set_video_device */

static const char *
set_video_device (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    if (!(*arg1)) {
	return "video_device cannot be empty";
    }

    d->device_name = ap_pstrdup(cmd->pool, arg1);

    return NULL;
}

/* }}} */
/* {{{ set_video_format */

static const char *
set_video_format (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    if (!(*arg1)) {
	return "video_format cannot be empty";
    }

    d->format = ap_pstrdup(cmd->pool, arg1);
    return NULL;
}

/* }}} */
/* {{{ set_video_input */

static const char *
set_video_input  (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    d->input = atoi(arg1);
    return NULL;
}

/* }}} */
/* {{{ set_video_width */

static const char *
set_video_width  (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    d->width = atoi(arg1);
    return NULL;
}

/* }}} */
/* {{{ set_video_height */

static const char *
set_video_height  (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    d->height = atoi(arg1);
    return NULL;
}

/* }}} */
/* {{{ set_video_jpeg_quality */

static const char *
set_video_jpeg_quality  (cmd_parms *cmd, video_dir_config *d, char *arg1)
{
    d->jpeg_quality = atoi(arg1);
    return NULL;
}

/* }}} */

/* {{{ various structures */

/* handler record */
static const handler_rec video_handlers[] =
{
    {"video", video_handler},
    {NULL}
};


/* configuration commands */
static command_rec video_cmds[] =
{
    {"video_device", set_video_device, NULL, ACCESS_CONF, TAKE1, "video device"},
    {"video_format", set_video_format, NULL, ACCESS_CONF, TAKE1, "video format"},
    {"video_input", set_video_input, NULL, ACCESS_CONF, TAKE1, "video input"},
    {"video_width", set_video_width, NULL, ACCESS_CONF, TAKE1, "video width"},
    {"video_height", set_video_height, NULL, ACCESS_CONF, TAKE1, "video height"},
    {"video_jpeg_quality", set_video_jpeg_quality, NULL, ACCESS_CONF, TAKE1, "jpeg quality"},
    {NULL}
};

module MODULE_VAR_EXPORT video_module = {
    STANDARD_MODULE_STUFF,
    video_init,                         /* module initializer     */
    video_create_dir_config,            /* per dir config creator */
    NULL,                               /* dir config merger */
    NULL,                               /* server config creator */
    NULL,                               /* server config merger*/
    video_cmds,                         /* command table */
    video_handlers,                     /* list of handlers */
    NULL,                               /* filename-to-URI translation */
    NULL,                               /* check_user_id */
    NULL,                               /* check auth */
    NULL,                               /* check access */
    NULL,                               /* type_checker */
    NULL,                               /* fixups */
    NULL,                               /* logger */
    NULL,                               /* header parser */
    NULL,                               /* child_init */
    NULL,                               /* child_exit */
    NULL                                /* post read-request */
};

/* }}} */
    
/* {{{ Emacs local variables */

/*
** Local variables:
** folded-file: t
** end:
*/

/* }}} */
